// SweepTriangle.cpp: 
// The code performs  Delaunay triangulation using sweep-line
// approach. There are three main fanctions need to be called for
// triangulation:
// 
// int InitSweepDT(long n, double* xarr, double* yarr)
// void FinalSweepDT()
// void SweepDT(long StartVertex)
//
// written by B. ZALIK in March and April 2003
// e-mail:zalik@uni-mb.si
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyTriangF.h"
#include "SweepTriangle.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// global variables

static VertexStruct* Vertices = NULL; // storrage of input vertices
long NoOfVertices = 0;                // number of vertices being triangulated  
static SimpleSkipList* AFLow = NULL;  // "advancing-front" implemented as the skip-list
static SimpleSkipList* AFHigh = NULL; // "bottom" advancing front implemented as the skip list
static long SwapCounter = 0;          // counts number of swaps


double C[3][3];                       // matrix for Delaunay criteria
BaseTriangle* TriangArray;            // array for storring triangles
long NextTriangPosition = 0;          // pointer to the array of triangles

//#define PRINTOUT    

int InitSweepDT(long n, double* xarr, double* yarr)
// initialisation of DT sweep-line algorithm:
// -n: number of points being triangulated
// - xarr, yarr: coordinate of the points
// Function creates the first triangle and initialise Low and
// High advancing front (AF). LowAF moves together with the
// sweep-line. Function also solves all the possible boundary
// cases at the initialisation (horizontal and vertical colliear 
// points). Regarding the special case meet, it returns the number
// of input points being already triangulated (input points are
// sorted according the y coordinate). In the expected case,
// three points is used during the initialization.
{  
  if (n < 2) return 0;

  SwapCounter = 0;
  double min = MAX_DOUBLE;
  double max = -MAX_DOUBLE;
  Vertices = new VertexStruct[n];
  for (long i = 0; i < n; i++)
  { Vertices[i].x = xarr[i];
    Vertices[i].y = yarr[i];
    Vertices[i].Index = i;
    if (xarr[i] < min) min = xarr[i];
    if (xarr[i] > max) max = xarr[i];
  }

  int iLo = 0;
  int iHi = n-1;
  QSort(iLo, iHi); // sort the points
 
  if (Equal(Vertices[0].y, Vertices[1].y))
  { if (Vertices[0].x < Vertices[1].x)
    { VertexStruct a = Vertices[0];
      Vertices[0] = Vertices[1];
      Vertices[1] = a;
    }
  }

  TriangArray = new BaseTriangle[2*n]; 
  NoOfVertices = n;

  long BottomFlag = 0;
  if (Equal(Vertices[0].y, Vertices[2].y))
  { if ((BottomFlag = BottomTriangles()) == 0)
      return NoOfVertices; // all vertices are collinear

    AFLow = new SimpleSkipList(NoOfVertices, 2, min, max);
    InitAFHigh();
    InitAFBottom(BottomFlag);

#ifdef PRINTOUT
 
FILE* f = fopen("VERTICES.TXT","wt");
for (i = 0; i < n; i++)
  fprintf(f, "%ld %lf %lf %ld \n", i, Vertices[i].x, 
                                      Vertices[i].y, 
                                      Vertices[i].Index);
fclose(f);
#endif


    return BottomFlag+1;
  }
  else // bottom points are not colinear
  { AFLow = new SimpleSkipList(NoOfVertices, 2, min, max);
 
    int returnNo = FirstTriangle();

    if (returnNo == VERTICALLY_COLINEAR)
    { return VerticallyColinearPoints();
    }  

    if (returnNo == COLINEAR)
      return InitCollinearPoints();

    if (returnNo == NoOfVertices) return 0;

    InitAFHigh();
    InitAF();


#ifdef PRINTOUT

FILE* f = fopen("VERTICES.TXT","wt");
for (i = 0; i < n; i++)
  fprintf(f, "%ld %lf %lf %ld \n", i, Vertices[i].x, 
                                      Vertices[i].y, 
                                      Vertices[i].Index);
fclose(f);
#endif

    return returnNo+1;
  }
}

void FinalSweepDT()
// function clean the memory after DT has been performed.
{
  if (Vertices != NULL)
  { delete[] Vertices;
    Vertices = NULL;
  }

  if (TriangArray != NULL)
  { delete[] TriangArray;
    TriangArray = NULL;
  }

  delete AFLow;
  AFLow = NULL;
  FreeAFHigh();
  AFLow = AFHigh = NULL;
}


void QSort(long iLo, long iHi) 
// quick sort of the array Vertices regarding y coordinate.
{
  long Lo, Hi;
  double Mid;
  VertexStruct T;

  Lo = iLo;
  Hi = iHi;
  Mid = Vertices[(Lo+Hi)/2].y;
  do
  {
    while (Vertices[Lo].y < Mid) Lo++;
    while (Vertices[Hi].y > Mid) Hi--;
    if (Lo <= Hi)
    {
      T = Vertices[Lo];
      Vertices[Lo] = Vertices[Hi];
      Vertices[Hi] = T;
      Lo++;
      Hi--;
    }
  }
  while (Lo <= Hi);

  if (Hi > iLo)
    QSort(iLo, Hi);
  if (Lo < iHi)
    QSort(Lo, iHi);
}

int InitAF()
// innitialisation of the advancing fronts (High and Low).
// Let us remember, that High AF belongs to the bottom convex hull
// and does not follow in general the sweep-line. Low AF on the 
// other hand follows the movements of the sweep-line and does not
// defines the part of the convex hull.
// function always returns 1
{ long i = TriangArray[0].i; // first triangle vertices
  long j = TriangArray[0].j;
  long k = TriangArray[0].k;


  if (Equal(Vertices[j].x, Vertices[i].x))
  { AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
 
    AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
    return 1;  
  }

  if (Equal(Vertices[j].x, Vertices[k].x))
  { AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
 
    AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
    return 1;  
  }

  if ( ((Vertices[j].x > Vertices[k].x) &&
        (Vertices[j].x < Vertices[i].x)) ||
       ((Vertices[j].x < Vertices[k].x) &&
        (Vertices[j].x > Vertices[i].x)))
  { 
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
        
    AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
    return 1;   
  }

  if (Vertices[i].x > Vertices[j].x) 
  {   
    if (Vertices[k].x < Vertices[i].x)
    { AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
      AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
      AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
      return 1;
    }
    else
    { AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);
      AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      AFHighAppendLeft(Vertices[j].x, Vertices[j].y, j, 0);
      AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, 0);
      return 1;
    }
  }

  if (Vertices[k].x < Vertices[i].x)
  { AFLow->Insert(Vertices[k].x, Vertices[k].y, k, 0);
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);

    AFHighAppendRight(Vertices[j].x, Vertices[j].y, j, 0);
    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    return 1;
  }
  else
  { AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->Insert(Vertices[j].x, Vertices[j].y, j, 0);

    AFHighAppendLeft(Vertices[k].x, Vertices[k].y, k, 0);
    AFHighAppendLeft(Vertices[i].x, Vertices[i].y, i, 0);
    AFHighAppendRight(Vertices[j].x, Vertices[j].y, j, 0);
    return 1;
  }
}


int InitCollinearPoints()
// function initialises triangles where the first i
// points is non-vertical and non-horizontal collinear
// it returns number of used points
{ InitAFHigh();
  long i = 0;
  double sg;
  while ((i+2 < NoOfVertices) && 
         ((sg = SignedArea(Vertices[i].x, Vertices[i].y,
                     Vertices[i+1].x, Vertices[i+1].y,
                     Vertices[i+2].x, Vertices[i+2].y)) == 0) )
    i++;

  long j;
  i+=2;
  if (i == NoOfVertices)
    return NoOfVertices;

  if (sg > 0)
  { for (j = 1; j < i; j++)
    { TriangArray[NextTriangPosition].i = i;
      TriangArray[NextTriangPosition].j = j;
      TriangArray[NextTriangPosition].k = j-1;

      TriangArray[NextTriangPosition].ij = NOINDEX;
      TriangArray[NextTriangPosition].jk = NOINDEX;

      if (j == 1)
        TriangArray[NextTriangPosition].ki = NOINDEX;
      else
      { TriangArray[NextTriangPosition].ki = NextTriangPosition-1;
        TriangArray[NextTriangPosition-1].ij = NextTriangPosition;
      }
    
      TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
      NextTriangPosition++;
    }

    if (Vertices[0].x > Vertices[1].x)
    { AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
      for (int k = 1; k < i-1; k++)
        AFLow->AppendAtTheLeft(Vertices[k].x, Vertices[k].y, k, k);

      AFLow->AppendAtTheLeft(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);
      AFLow->AppendAtTheLeft(Vertices[i].x, Vertices[i].y, i, i-2);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, 0);

      return i+1;
    }

    if (Vertices[i].x > Vertices[i-1].x)
    { AFLow->Insert(Vertices[i].x, Vertices[i].y, i, i-2);
      AFLow->AppendAtTheLeft(Vertices[0].x, Vertices[0].y, 0, 0);
      

      for (int k= 0; k < i-1; k++)
        AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, k);

      AFHighAppendRight(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);
      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, i-2);
      return i+1;
    }

    if (Vertices[i].x < Vertices[0].x)
    { 
      AFLow->Insert(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);
      AFLow->AppendAtTheLeft(Vertices[i].x, Vertices[i].y, i, i-2);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      for (int k= 0; k < i-1; k++)
        AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, k);

      AFHighAppendRight(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);

      return i+1;
    }

    AFLow->Insert(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);
    AFLow->Insert(Vertices[i].x, Vertices[i].y, i, i-2);
    AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
    AFLow->ChangeTriangleIndex(0,0);
    AFLow->ChangeTriangleIndex(i,i-2);
    AFLow->ChangeTriangleIndex(i-1,i-2);

    for (int k= 0; k < i-1; k++)
      AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, k);

     AFHighAppendRight(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);
    return i+1;
  }
  else
  { for (j = 1; j < i; j++)
    { TriangArray[NextTriangPosition].i = i;
      TriangArray[NextTriangPosition].j = j-1;
      TriangArray[NextTriangPosition].k = j;
      TriangArray[NextTriangPosition].ki = NOINDEX;
      TriangArray[NextTriangPosition].jk = NOINDEX;

      if (j == 1)
        TriangArray[NextTriangPosition].ij = NOINDEX;
      else
      { TriangArray[NextTriangPosition].ij = NextTriangPosition-1;
        TriangArray[NextTriangPosition-1].ki = NextTriangPosition;
      }
    
      TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
      NextTriangPosition++;
    } 

    if (Vertices[0].x < Vertices[1].x)
    { 

      AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
      AFLow->AppendAtTheLeft(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);

      for (int k = i-1; k > 0; k--)
        AFLow->AppendAtTheLeft(Vertices[k].x, Vertices[k].y, k, k);

      AFLow->AppendAtTheLeft(Vertices[0].x, Vertices[0].y, 0, 0);

      AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, 0);
      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      return i+1;
    }


    if (Vertices[i].x < Vertices[i-1].x)
    { AFLow->Insert(Vertices[i].x, Vertices[i].y, i, 0);
      AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);

      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, i-2);
      for (int k= i-1; k > 0 ; k--)
        AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, k-1);

      AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, 0);
      return i+1;
    }

    if (Vertices[i].x > Vertices[0].x)
    { AFLow->Insert(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);
      AFLow->Insert(Vertices[i].x, Vertices[i].y, i, i-2);

      for (int k= i-1; k > 0 ; k--)
        AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, k-1);

      AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, 0);
      AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
      return i+1;
    }
  
    AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
    AFLow->AppendAtTheLeft(Vertices[i].x, Vertices[i].y, i, 0);
    AFLow->AppendAtTheLeft(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);

    AFLow->ChangeTriangleIndex(0,0);
    AFLow->ChangeTriangleIndex(i,0);
    AFLow->ChangeTriangleIndex(i-1,i-2);

    for (int k= i-1; k > 0 ; k--)
      AFHighAppendRight(Vertices[k].x, Vertices[k].y, k, k-1);

    AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, 0);
    return i+1;
  }
}

void InitAFBottom(long n )
// function initialises the advancing fronts in the special
// case when more that two input vertices are colinear.
// in this case, more that one triangle is generated at the
// initialisation, because all the bottom colinear vertices
// are connected with the first non-colinear vertex. 
{ 

  if ((Vertices[0].x <= Vertices[n].x) && 
      (Vertices[n].x <= Vertices[n-1].x) )
  {
    if (Equal(Vertices[0].x, Vertices[n].x))
    {
      AFLow->Insert(Vertices[n].x, Vertices[n].y, n, n-2);
      AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
      AFLow->Insert(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
      AFLow->UpdateTIndex(Vertices[n].x, n, n-1, n-2);
      AFLow->UpdateTIndex(Vertices[0].x, 0, n, 0);
    }
    else
    { if (Equal(Vertices[n].x, Vertices[n-1].x))
      {
        AFLow->Insert(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
        AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
        AFLow->Insert(Vertices[n].x, Vertices[n].y, n, n-2);
        AFLow->UpdateTIndex(Vertices[n].x, n, n-1, n-2);
        AFLow->UpdateTIndex(Vertices[0].x, 0, n, 0);
      }
      else
      {
        AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
        AFLow->AppendAtTheRight(Vertices[n].x, Vertices[n].y, n, n-2);
        AFLow->AppendAtTheRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
        AFLow->UpdateTIndex(Vertices[n].x, n, n-1, n-2);
        AFLow->UpdateTIndex(Vertices[0].x, 0, n, 0);
      }
    }

    AFHighAppendRight(Vertices[0].x, Vertices[0].y, 0, NOINDEX);
    AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
  }
  else // AF contains only two vertices
  { if (Vertices[n].x < Vertices[0].x)
    { AFLow->Insert(Vertices[n].x, Vertices[n].y, n, n-2);
      AFLow->AppendAtTheRight(Vertices[n-1].x, Vertices[n-1].y, n-1, n-2);
      AFLow->UpdateTIndex(Vertices[n-1].x, n, n-1, n-2); 
      
      AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
//      AFHighAppendLeft(Vertices[0].x, Vertices[0].y, 0, NOINDEX);
//      AFHighAppendLeft(Vertices[n].x, Vertices[n].y, n, NOINDEX);
      AFHighAppendLeft(Vertices[0].x, Vertices[0].y, 0, 0);
      AFHighAppendLeft(Vertices[n].x, Vertices[n].y, n, 0);
    } 
    else
    { AFLow->Insert(Vertices[0].x, Vertices[0].y, 0, 0);
      AFLow->AppendAtTheRight(Vertices[n].x, Vertices[n].y, n, 0);
      AFLow->UpdateTIndex(Vertices[n].x, 0, n, 0); 

      AFHighAppendLeft(Vertices[0].x, Vertices[0].y, 0, NOINDEX);
//      AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NOINDEX);
//      AFHighAppendRight(Vertices[n].x, Vertices[n].y, n, NOINDEX);
      AFHighAppendRight(Vertices[n-1].x, Vertices[n-1].y, n-1, NextTriangPosition-1);
      AFHighAppendRight(Vertices[n].x, Vertices[n].y, n, NextTriangPosition-1);

    }
  }
}

int VerticallyColinearPoints()
// the last special case of the initialisation. This function is called
// when more than two first vertices are vertically collinear.
// also in this case, more triangles are generated at the initialisation,
// and both advancing fronts are initialised adequatelly.
// function returns the number of used input points (all vertically
// collinear plus one non-collinear.
{
  InitAFHigh();
  long i = 1;

  // count the number for the first vertically collinear vertices
  while (Equal(Vertices[i-1].x,Vertices[i].x) && (i < NoOfVertices))
    i++;

  // is the triangulation possible?
  if (i == NoOfVertices) return NoOfVertices;

  long t;
  NextTriangPosition = 0; 
  for (t = 1; t < i; t++) // construct triangles
  { 
    TriangArray[NextTriangPosition].i = i;
    TriangArray[NextTriangPosition].j = t-1;
    TriangArray[NextTriangPosition].k = t;

    TriangArray[NextTriangPosition].ij = 
    TriangArray[NextTriangPosition].jk =
    TriangArray[NextTriangPosition].ki = NONE;

    if (Vertices[i].x < Vertices[0].x)
    {
      TriangArray[NextTriangPosition].i = t;
      TriangArray[NextTriangPosition].k = i;

      if (t > 1)    
      { TriangArray[NextTriangPosition].jk = NextTriangPosition - 1;
        TriangArray[NextTriangPosition-1].ki = NextTriangPosition;
        TriangArray[NextTriangPosition].ij =
        TriangArray[NextTriangPosition].ki = NONE;
      }
    }
    else
    { if (t > 1)    
      { TriangArray[NextTriangPosition].ij = NextTriangPosition - 1;
        TriangArray[NextTriangPosition-1].ki = NextTriangPosition;
        TriangArray[NextTriangPosition].jk =
        TriangArray[NextTriangPosition].ki = NONE;
      }
    }
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
    NextTriangPosition++;
  }

  AFLow->Insert(Vertices[i].x, Vertices[i].y, i, i-2);
  AFLow->Insert(Vertices[i-1].x, Vertices[i-1].y, i-1, i-2);

  if (Vertices[i].x < Vertices[0].x)
  { AFHighAppendLeft(Vertices[i].x, Vertices[i].y, i, 0);
    for (t = 0; t < i; t++)
      AFHighAppendRight(Vertices[t].x, Vertices[t].y, t, t);
  }
  else
  { AFHighAppendRight(Vertices[i].x, Vertices[i].y, i, 0);
    for (t = 0; t < i; t++)
      AFHighAppendLeft(Vertices[t].x, Vertices[t].y, t, t);
  }

  return i+1;
}

int FirstTriangle()
// function construct the first triangle and orients it 
// in counterclock-wise direction. It detects also the possible
// boundary cases (collinear vertices).
{

  if (Equal(Vertices[0].x, Vertices[1].x) &&
      Equal(Vertices[0].x, Vertices[2].x))  
    return VERTICALLY_COLINEAR; // points are vertically colinear
 
  // let us eliminate the same points at the beginning
  long i = 1;
  while ( TheSamePoints2D(Vertices[0].x,Vertices[0].y,
                          Vertices[i].x,Vertices[i].y) && 
          (i < NoOfVertices)) i++;

  if (i == NoOfVertices) return NoOfVertices;

  long j = i+1;
  while ( TheSamePoints2D(Vertices[i].x,Vertices[i].y,
                          Vertices[j].x,Vertices[j].y) && 
          (j < NoOfVertices)) j++;

  if (i == NoOfVertices) return NoOfVertices;

  // there exist three different points

  NextTriangPosition = 0; 
  TriangArray[NextTriangPosition].i = j;
  TriangArray[NextTriangPosition].j = 0;
  TriangArray[NextTriangPosition].k = i;

  double sa = SignedArea(Vertices[0].x, Vertices[0].y, 
                         Vertices[j].x, Vertices[j].y,
                         Vertices[i].x, Vertices[i].y);

  if (Equal(sa,0))
    return COLINEAR;


  if (sa < 0)
  { TriangArray[NextTriangPosition].i = i;
    TriangArray[NextTriangPosition].k = j;
  }
  TriangArray[NextTriangPosition].ij = 
  TriangArray[NextTriangPosition].jk =
  TriangArray[NextTriangPosition].ki = NONE;
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
  NextTriangPosition++;
  return j;
  


/*


  NextTriangPosition = 0; 
  TriangArray[NextTriangPosition].i = 2;
  TriangArray[NextTriangPosition].j = 0;
  TriangArray[NextTriangPosition].k = 1;

  double sa = SignedArea(Vertices[0].x, Vertices[0].y, 
                         Vertices[2].x, Vertices[2].y,
                         Vertices[1].x, Vertices[1].y);

  if (Equal(sa,0)) // points are vertically colinear
  { return VERTICALLY_COLINEAR;
  }
  else
  { if (sa < 0)
    { TriangArray[NextTriangPosition].i = 1;
      TriangArray[NextTriangPosition].k = 2;
    }
    TriangArray[NextTriangPosition].ij = 
    TriangArray[NextTriangPosition].jk =
    TriangArray[NextTriangPosition].ki = NONE;
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
    NextTriangPosition++;
    return 1;
  }
*/
}



int BottomTriangles()
// Function creates the innitial triangles in the case of
// more vertices being located on the first scan-line
{ long i = 1;
  long n = 0;

  while ((i < NoOfVertices) && Equal(Vertices[i].y, Vertices[i-1].y))
  { i++;
    n++;
  }
  if (n == NoOfVertices)
    return 0;

  long j;
  VertexStruct VS;
  for (i = 0; i < n; i++)
    for (j = i+1; j < n+1; j++)
    { if (Vertices[j].x < Vertices[i].x)
      { VS = Vertices[j];
        Vertices[j] = Vertices[i];
        Vertices[i] = VS;
      }
    }

  n++;
  NextTriangPosition = 0;
  for (i = 1; i < n; i++)
  { TriangArray[NextTriangPosition].i = n;
    TriangArray[NextTriangPosition].j = i;
    TriangArray[NextTriangPosition].k = i-1;

    TriangArray[NextTriangPosition].ij = NOINDEX;
    TriangArray[NextTriangPosition].jk = NOINDEX;
 //   TriangArray[NextTriangPosition].ij = i;
 //   TriangArray[NextTriangPosition].ki = i;
    if (i == 1)
      TriangArray[NextTriangPosition].ki = NOINDEX;
    else
    { TriangArray[NextTriangPosition].ki = NextTriangPosition-1;
      TriangArray[NextTriangPosition-1].ij = NextTriangPosition;
    }
    
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
    NextTriangPosition++;
  }
//  TriangArray[NextTriangPosition-1].ij = NOINDEX;
  return n;
}


void SweepDT(long StartVertex)
// this is the main triangulation function performing DT
// using sweep-line approach. It works as follows:
// - for a point on the sweep-line, a vertical projection
// on the advancing front is made If there is more than 
// one vertex on the sweep-line, all of them are treated
// in the same way, and the order of their applying is arbitratry.
// - the projection on the advancing from (it is implemented by the
// skip-list) give us the information of the triangle which touches
// the edge of the advancing front. For cases are possible:
// 1. Projection of the point fails somewhere between the left and
//    the right border of the advancing front (the most frequent 
//    case.
// 2. Projection misses the advancing front from the left side
// 3. Projection misses the advancing front from the right side
// 4. Considered point is located at an edge of the advancing front.
// This function dispatch these four cases and calls the corresponding
// functions.
// - at the end, the advancing front is triangulated towards the 
// upper convex hull.
// Function accept one parameter telling how many points have been
// already triangulated during the initialization.
{ long i;
  DoubleLinkedList* LeftAF;
  DoubleLinkedList* RightAF;

//NoOfVertices = 320;
//NoOfVertices = 4;
#ifdef PRINTOUT
AFLow->PrintEntries();
AFLow->PrintAFLow();
PrintTriangulation();
#endif

  for (i = StartVertex; i < NoOfVertices; i++)
  { 


#ifdef PRINTOUT

if (i == 7)
{
int  uuu = 880;
}

FILE* bf = fopen("i.txt", "wt");
fprintf(bf, "%ld\n", i);
fclose(bf);
#endif



      switch (AFLow->ReturnInterval(Vertices[i].x,
                                    Vertices[i].y, 
                                    &LeftAF, 
                                    &RightAF))
      { case MIDDLE: 
          SolveMiddleCase(i, LeftAF, RightAF); 
          break;
        case LEFT: 
          SolveLeftCase(i); 
          break;
        case RIGHT: 
          SolveRightCase(i);
          break;
        case ON_EDGE:
          SolveOnEDGE(i, LeftAF, RightAF); // se napaka pri sosedih
         break;           
      }


#ifdef PRINTOUT
AFLow->PrintEntries();
AFLow->PrintAFLow();
PrintTriangulation();
PrintDCLList();
#endif
 
  }
  CloseTriangles();
}


int CloseTriangles()
// function closes the triangles in the advancing front to
// form the upper convex hull of the input set of the points
// Function if not optimized as it is called only once and as 
// usually advancing front does contain only a small pertence
// of the input points. Function always returns 1.
{
  int ContinueFlag = 1;
  int TInd1, TInd2;
  DoubleLinkedList* w1;
  DoubleLinkedList* w2;
  DoubleLinkedList* w3;
  DoubleLinkedList* RightWatch = AFLow->ReturnRight();

  while (ContinueFlag)
  { ContinueFlag = 0;
    w1 = AFLow->ReturnLeft();
    if (w1 == NULL) return 1;
    w2 = w1->RPointer;
    if (w2 == NULL) return 1;
    w3 = w2->RPointer;

    while (w3 != RightWatch->RPointer)
    { if (SignedArea(w1->key, w1->y, w2->key, w2->y, w3->key, w3->y) > 0)
      { TInd1 = w1->TriangleIndex;
        TInd2 = w2->TriangleIndex;

        CloseTriangle(w1, w2, w3);
        AFLow->DeleteElement(w2, NextTriangPosition);

        CheckDT(NextTriangPosition, TInd1);
        CheckDT(NextTriangPosition, TInd2);
        NextTriangPosition++;
        ContinueFlag = 1;
        w2 = w3;
        w3 = w3->RPointer;
      }
      else
      { w1 = w2;
        w2 = w3;
        w3 = w3->RPointer;
      }
    }
  }
  return 1;
}


void SetNeighbour(long T1, long v1, long v2, long TIndex)
// Function sets the new neighbour TIndex of triangle T1 over the
// vertices v1 v2.
{ if ( ((TriangArray[T1].i == v1) && (TriangArray[T1].j == v2)) ||
       ((TriangArray[T1].i == v2) && (TriangArray[T1].j == v1)))
    TriangArray[T1].ij = TIndex;
  else
  { if ( ((TriangArray[T1].j == v1) && (TriangArray[T1].k == v2)) ||
         ((TriangArray[T1].j == v2) && (TriangArray[T1].k == v1)))
      TriangArray[T1].jk = TIndex;
    else
     TriangArray[T1].ki = TIndex;
  }
}


void InsertNewTriangle(long NewVertexIndex, DoubleLinkedList* lt, DoubleLinkedList* rt)
// Function constructs a new triangle.
// - NewVertexIndex: index of the vertex being located at the current sweep-lin
// - lt: pointer in the advancing front on the left vertex of the
//       vertical projection of NewVertexIndex
// - rt: pointer to the right vertex projection 
{ long OldT = lt->TriangleIndex;
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = rt->VertexNo;
  long k = TriangArray[NextTriangPosition].k = lt->VertexNo;

  TriangArray[NextTriangPosition].ij = 
  TriangArray[NextTriangPosition].ki = NONE;
  TriangArray[NextTriangPosition].jk = OldT;

  SetNeighbour(OldT, k, j,NextTriangPosition);
  CheckBorderFlag(OldT);

  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}

void CloseNewTriangleLeft(long NewVertexIndex, 
                          DoubleLinkedList* LeftP, 
                          DoubleLinkedList* NextLeftP)
// construction of the new triangle when NewVertexIndex misses
// the advanding front from the left side. LeftP and the NextLeftP
// are pointers to the first left and the next vertex from the left
// of the advancing front.
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = LeftP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextLeftP->VertexNo;

  TriangArray[NextTriangPosition].ij = LeftP->TriangleIndex;
  TriangArray[NextTriangPosition].ki = NOINDEX;
  TriangArray[NextTriangPosition].jk = NextLeftP->TriangleIndex;

  SetNeighbour(LeftP->TriangleIndex, j, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(LeftP->TriangleIndex);

  SetNeighbour(NextLeftP->TriangleIndex, j, k, NextTriangPosition);
  CheckBorderFlag(NextLeftP->TriangleIndex);
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}


void CloseNewTriangleLeftMiddle(long NewVertexIndex, 
                          DoubleLinkedList* LeftP, 
                          DoubleLinkedList* NextLeftP)
// simmilar us upper function but this time, the vertical 
// projection of the NewVertexIndex misses the advancing from
// from the right side
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = LeftP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextLeftP->VertexNo;

  TriangArray[NextTriangPosition].ij = LeftP->TriangleIndex;
  TriangArray[NextTriangPosition].jk = NextLeftP->TriangleIndex;
  TriangArray[NextTriangPosition].ki = NOINDEX;

  SetNeighbour(NextLeftP->TriangleIndex, k, j, NextTriangPosition);
  CheckBorderFlag(NextLeftP->TriangleIndex);

  SetNeighbour(LeftP->TriangleIndex, j, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(LeftP->TriangleIndex);
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}


void CheckBorderFlag(int TIndex)
// function determines whether triangle TIndex determines
// the border of advancing fronts (both, upper and lower)
// and set the corresponding flag.
{ 
  if ((TriangArray[TIndex].ij != NOINDEX) &&
      (TriangArray[TIndex].jk != NOINDEX) &&
      (TriangArray[TIndex].ki != NOINDEX))
    TriangArray[TIndex].BorderTriangleFlag = FALSE;
  else
    TriangArray[TIndex].BorderTriangleFlag = TRUE;
}


void CloseNewTriangleLeftDown(long NewVertexIndex, 
                              long TIndex,
                              DoubleLinkedList* LeftP, 
                              DoubleLinkedList* NextLeftP)
// this function constructs the triangles along the left 
// side of the lower convex hull and extents it adequatelly
// - NewVertexIndex: index of inserted vertex
// - TIndex: index of the neighbouring triangle along the edge ki
// - LeftP, NextLeftP;  pointer to the left and the next left
//    vertex of the lower convex hull front.
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = LeftP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextLeftP->VertexNo;

  TriangArray[NextTriangPosition].ij = NOINDEX;
  TriangArray[NextTriangPosition].ki = TIndex;
  TriangArray[NextTriangPosition].jk = LeftP->TriangleIndex;
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;

  if (TIndex != NOINDEX)
  { SetNeighbour(TIndex, TriangArray[NextTriangPosition].k, 
                         TriangArray[NextTriangPosition].i, 
                         NextTriangPosition); //!!!!
    CheckBorderFlag(TIndex);
  }

  SetNeighbour(LeftP->TriangleIndex, k, j,NextTriangPosition);
  CheckDT(LeftP->TriangleIndex, NextTriangPosition);


  CheckBorderFlag(LeftP->TriangleIndex);
}



void CloseNewTriangleRightDown(long NewVertexIndex, 
                              long TIndex,
                              DoubleLinkedList* RightP, 
                              DoubleLinkedList* NextRightP)
// simmilar as above, but this time triangle is constructed
// on the right side of the lower convex hull
{ 
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = RightP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = NextRightP->VertexNo;

  TriangArray[NextTriangPosition].ij = TIndex;
  TriangArray[NextTriangPosition].jk = RightP->TriangleIndex ;
  TriangArray[NextTriangPosition].ki = NOINDEX;
  CheckBorderFlag(NextTriangPosition);

  if (TIndex != NOINDEX) 
  { 
    SetNeighbour(TIndex, TriangArray[NextTriangPosition].j, 
                         TriangArray[NextTriangPosition].i, 
                         NextTriangPosition); //!!!!
    CheckBorderFlag(TIndex);
  }

  SetNeighbour(RightP->TriangleIndex, k, j,NextTriangPosition);
  CheckDT(RightP->TriangleIndex, NextTriangPosition);


  CheckBorderFlag(RightP->TriangleIndex);
}


void CloseNewTriangleRight(long NewVertexIndex, 
                           DoubleLinkedList* RightP, 
                           DoubleLinkedList* NextRightP)
// this cuntion construct a triangle which is on the right side 
// of the inserted vertex NewVertexIndex. This triangle has an
// sharp angle with vertices NewVertexIndex, RightP, NextRightP
// (pointers in the AF).
{
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = NextRightP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = RightP->VertexNo;

  TriangArray[NextTriangPosition].ij = NOINDEX; 
  TriangArray[NextTriangPosition].ki = NextTriangPosition - 1;
  long TIndex2 = RightP->TriangleIndex;
  TriangArray[NextTriangPosition].jk = TIndex2;
  CheckBorderFlag(NextTriangPosition);

//  TriangArray[NextTriangPosition - 1].ij = NextTriangPosition;
  SetNeighbour(NextTriangPosition - 1, k, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(NextTriangPosition - 1);

  SetNeighbour(TIndex2, k, j,NextTriangPosition);
  CheckBorderFlag(TIndex2);
}


void CloseNewTriangleRightMiddle(long NewVertexIndex, 
                           DoubleLinkedList* RightP, 
                           DoubleLinkedList* NextRightP)
// Function close a triangle in the right direction of the 
// projection of NewVertexIndex
{
  TriangArray[NextTriangPosition].i = NewVertexIndex;
  long j = TriangArray[NextTriangPosition].j = NextRightP->VertexNo;
  long k = TriangArray[NextTriangPosition].k = RightP->VertexNo;

  TriangArray[NextTriangPosition].ij = NOINDEX; 
  TriangArray[NextTriangPosition].jk = RightP->TriangleIndex;
  long Ti = RightP->LPointer->TriangleIndex;
  TriangArray[NextTriangPosition].ki = Ti;

  SetNeighbour(Ti, k, NewVertexIndex, NextTriangPosition);
  CheckBorderFlag(Ti);

  SetNeighbour(RightP->TriangleIndex, k, j, NextTriangPosition);
  CheckBorderFlag(RightP->TriangleIndex);
}


int SolveOnEDGE(long nv, DoubleLinkedList* Left, 
                        DoubleLinkedList* Right)
// in this case the point is on the edge of the advancing front.
// Therefore, the triangle which is touching the advancing front
// along edges Left and Right must be split
{ long t1 = Left->TriangleIndex;
  long v1 = Left->VertexNo;
  long v2 = Right->VertexNo;

  int Orientation;
  int flag = 0;
  while ((( TriangArray[t1].i != v1) &&  
          ( TriangArray[t1].j != v2)) && (flag != 3))           
  { ShiftVerticesRight(&TriangArray[t1]);   
    flag++;
  }

  if (flag < 3)
    Orientation = 1;  // left orientation ijk
  else
  { flag = 0;
    Orientation = 0;  // left orientation kij
    while ((( TriangArray[t1].k != v2) &&  
             ( TriangArray[t1].i != v1)) && (flag != 3))           
    { ShiftVerticesRight(&TriangArray[t1]);   
      flag++;
    }
  }

  long i, j, k, n_ij, n_jk, n_ki;
  if (Orientation) // triangle is ijk
  { i = TriangArray[t1].i;
    j = TriangArray[t1].j;
    k = TriangArray[t1].k;
    n_jk = TriangArray[t1].jk;
    n_ki = TriangArray[t1].ki;

    TriangArray[t1].j = nv;
    TriangArray[t1].jk = NextTriangPosition;
    
    TriangArray[NextTriangPosition].i = nv;
    TriangArray[NextTriangPosition].j = j;
    TriangArray[NextTriangPosition].k = k;
    TriangArray[NextTriangPosition].ij = NOINDEX;
    TriangArray[NextTriangPosition].jk = n_jk;
    TriangArray[NextTriangPosition].ki = t1;
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;


//    SetNeighbour(OldT, k, j,NextTriangPosition);
    if (n_jk != NOINDEX)
       ReplaceNeighbour1(n_jk, t1, NextTriangPosition);

 
    AFLow->Insert(Vertices[nv].x,
                  Vertices[nv].y, nv, NextTriangPosition);
    AFLow->UpdateTIndex(Vertices[i].x, i, nv, t1);

    CheckDT(t1, n_ki);
    CheckDT(n_jk, NextTriangPosition);
  }
  else // orientation is kij
  { i = TriangArray[t1].i;
    j = TriangArray[t1].j;
    k = TriangArray[t1].k;

    n_ij = TriangArray[t1].ij;
    n_jk = TriangArray[t1].jk;

    TriangArray[t1].i = nv;
    TriangArray[t1].ij = NextTriangPosition;

    TriangArray[NextTriangPosition].i = i;
    TriangArray[NextTriangPosition].j = j;
    TriangArray[NextTriangPosition].k = nv;
    TriangArray[NextTriangPosition].ij = n_ij;
    TriangArray[NextTriangPosition].jk = t1;
    TriangArray[NextTriangPosition].ki = NOINDEX;
    TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;

    if (n_jk != NOINDEX)
       ReplaceNeighbour1(n_jk, t1, NextTriangPosition);

    AFLow->Insert(Vertices[nv].x,
                  Vertices[nv].y, nv, NextTriangPosition);
    AFLow->UpdateTIndex(Vertices[i].x, i, nv, t1);

    CheckDT(t1, n_jk);
    CheckDT(n_ij, NextTriangPosition);
  } 
  NextTriangPosition++;
  return 1;
}

int SolveMiddleCase(long i, DoubleLinkedList* Left,
                            DoubleLinkedList* Right)
// The most usual case, the vertical projection of the new
// vertex i fails between Left and Right vertices of the 
// advancing front. Function always returns 1.
{ double x = Vertices[i].x;
  double y = Vertices[i].y;
  long Pair = Left->TriangleIndex;
  InsertNewTriangle(i, Left, Right);


  AFLow->Insert(x,y,i,NextTriangPosition);    

  CheckDT(Pair, NextTriangPosition);
  NextTriangPosition++;

  int MiddleFlag = 1;

  DoubleLinkedList* NewRight = GoRight(x, y, i, Right);
  DoubleLinkedList* NewLeft = GoLeft(x, y, i, Left);

  SolveGapRight(NewRight);

  SolveGapLeft(NewLeft);

  return 1;
}


int SolveGapRight(DoubleLinkedList* middleAF)
// The right gap in the advancing front is tried to be filled
// the condition for that is, that the angle between the right
// edge pointed by middleAF and the next edge in the advancing front is
// larger that 135 degrees. 
{ DoubleLinkedList* backAF = middleAF->LPointer;
  DoubleLinkedList* nextAF = middleAF->RPointer;

  double dx, dy;
  
  if ((nextAF->RPointer != NULL) && (nextAF->y - middleAF->y < 0))
  { dx = nextAF->key - backAF->key; 
    if (Equal(dx, 0))
    { dx = 1;
      dy = 1000; 
    }
    else
      dy = backAF->y - nextAF->y; // it is always positive
    if ( dy/dx > GAP_BORDER )
    { // cover the right gap
      // determine chain
      DoubleLinkedList *C1 = middleAF;
      DoubleLinkedList *C2 = nextAF;
      DoubleLinkedList *C3 = nextAF->RPointer;

      int s = 0;   
      while ((C3->RPointer != NULL) &&
             (SignedArea(C1->key, C1->y,C2->key,C2->y, C3->key, C3->y) > 0)) 
      { C1 = C2; C2 = C3; 
        C3 = C3->RPointer;
        s = 1;
      }
      if (s && (C3->LPointer != NULL)) 
        SewTheGap(middleAF, C2,0);      
    }
  }
  return 1;
}

int SolveGapLeft(DoubleLinkedList* middleAF)
// simmilar as above but this time, the gap on the left side 
// is filled
{ 
  DoubleLinkedList* backAF = middleAF->RPointer;
  DoubleLinkedList* nextAF = middleAF->LPointer;

  double dx, dy;
  
  if ((nextAF->LPointer != NULL) && (nextAF->y - middleAF->y < 0))
  { dx = backAF->key - nextAF->key; 
    if (Equal(dx, 0))
    { dx = 1;
      dy = 1000; 
    }
    else
      dy = backAF->y - nextAF->y; // it is always positive
    if ( dy/dx > GAP_BORDER )
    { // cover the right gap
      // determine chain
      DoubleLinkedList *C1 = middleAF;
      DoubleLinkedList *C2 = nextAF;
      DoubleLinkedList *C3 = nextAF->LPointer; 

      int s = 0;      
      while ((C3->LPointer != NULL) &&
             (SignedArea(C1->key, C1->y,C2->key, C2->y, C3->key, C3->y) < 0) )
      { C1 = C2; C2 = C3; 
        C3 = C3->LPointer;
        s = 1;
      }
      if (s && (C3->LPointer != NULL))
        SewTheGap(C2, middleAF,1);      
    }
  }
 return 1;
}



int SewTheGap(DoubleLinkedList* LBorder, 
              DoubleLinkedList* RBorder,
              int LeftSide)
// this fanction sews the gap determined by previous functions.
// LBorder and RBorder are pointers in the advancing front and
// between them the advancing front will be filled.
// -LeftSide (this parameter is not used; it is the remaining of
// the previous attempts. Function always returns 1.
{ 
  long Ti1, Ti2;
  DoubleLinkedList* w1 = LBorder;
  DoubleLinkedList* w2 = w1->RPointer;
  DoubleLinkedList* w3;
  while ((w2 != RBorder) && (w2->y < w1->y))
  { w1 = w2; // find the bottom vertex and store it in w1
    w2 = w2->RPointer;
  }
  if (w1 != LBorder) // we moved for at least one vertex in the chain
  { w2 = w1;
    w1 = w1->LPointer;
  }

  if (Equal(w1->y, w2->y))
  {  // solve colinear bottom vertices
    return 1;
  }
  else
  { if (w2 != RBorder)
      w3 = w2->RPointer;
    else
      w3 = RBorder;
   
    
    while ((w1 != LBorder) || (w3 != RBorder))
    { Ti1 = w1->TriangleIndex;
      Ti2 = w2->TriangleIndex;
      CloseTriangle(w1, w2, w3);
 
      AFLow->DeleteElement(w2, NextTriangPosition);

      CheckDT(NextTriangPosition, Ti1);
      CheckDT(NextTriangPosition, Ti2);


      NextTriangPosition++;

      if ((w1 == LBorder) && (w3 == RBorder))
      { ; 
      }
      else
      { if ((w1 == LBorder) && (w3 != RBorder))
        { w2 = w3;
          w3 = w3->RPointer;
        }
        else
        { if ((w1 != LBorder) && (w3 == RBorder))
          { w2 = w1;
            w1 = w1->LPointer;
          }
          else 
          { if (w1->LPointer->y > w3->RPointer->y)
            { w2 = w1;
              w1 = w1->LPointer;
            }
            else
            { w2 = w3;
              w3 = w3->RPointer;
            }
          }
        }
      }
    }
  }
  return 1;
}


void CloseTriangle(DoubleLinkedList* lft, 
                   DoubleLinkedList* middle, 
                   DoubleLinkedList* rght)
// this function make a triangle between lft, middle, and rght 
// vertices of the advancing front-
{ 
  long i =TriangArray[NextTriangPosition].i = lft->VertexNo;
  long j = TriangArray[NextTriangPosition].j = rght->VertexNo;
  long k = TriangArray[NextTriangPosition].k = middle->VertexNo;

  TriangArray[NextTriangPosition].ij = NONE;
  TriangArray[NextTriangPosition].ki = lft->TriangleIndex;
  TriangArray[NextTriangPosition].jk = middle->TriangleIndex; 

  SetNeighbour(middle->TriangleIndex, j, k,NextTriangPosition);
  CheckBorderFlag(middle->TriangleIndex);

  SetNeighbour(lft->TriangleIndex, k, i, NextTriangPosition);
  CheckBorderFlag(lft->TriangleIndex);
  TriangArray[NextTriangPosition].BorderTriangleFlag = TRUE;
}


DoubleLinkedList* GoRight(double x, double y, long i, DoubleLinkedList* rp)
// -(x,y) coordinates of the latest point inserted
// -i index of inserted point
// -rp: right pointer in AF regarding inserted point
// returns 1 is move to the right has been done (in this case rp is erased
// or 0 if the move has not been done
{
  DoubleLinkedList* AFWalker = rp;
  DoubleLinkedList* NextRightVertex= AFLow->ReturnNextRight();

  int WorkFlag = 1;
  long TInd;
  int CleanRP = 0;
  while ((NextRightVertex != NULL) && WorkFlag)
  { WorkFlag = 0;
    if ((AFWalker->y < y)&& 
        (AFWalker->y < NextRightVertex->y) &&
        (ScalarProduct(x, y, 
                       AFWalker->key, 
                       AFWalker->y,
                       NextRightVertex->key, 
                       NextRightVertex->y) > 0))
    { CloseNewTriangleRightMiddle(i, AFWalker, NextRightVertex);
      TInd = AFWalker->TriangleIndex;
      AFLow->DeleteElement(AFWalker, NextTriangPosition);
      WorkFlag = 1;
      CleanRP = 1;
    }

    if (WorkFlag)
    { CheckDT(NextTriangPosition, TInd);
      CheckDT(NextTriangPosition, TriangArray[NextTriangPosition].jk);

      NextTriangPosition++;
      AFWalker = NextRightVertex;
      NextRightVertex = AFLow->ReturnNextRight();
    }
  }
  return AFWalker;
}



DoubleLinkedList* GoLeft(double x, double y, long i, DoubleLinkedList* lp)
// -(x,y) coordinates of the latest point inserted
// -i index of inserted point
// -lp: left pointer in AF regarding inserted point
{
  DoubleLinkedList* AFWalker = lp;
  DoubleLinkedList* NextLeftVertex= AFLow->ReturnNextLeft();

  int WorkFlag = 1;
  int CleanRP = 0;
  long TInd;
  while ((NextLeftVertex != NULL) && WorkFlag)
  { WorkFlag = 0;

    if ((AFWalker->y < y)&& 
        (AFWalker->y < NextLeftVertex->y) &&
        (ScalarProduct(x, y, 
                       AFWalker->key, 
                       AFWalker->y,
                       NextLeftVertex->key, 
                       NextLeftVertex->y) > 0))
    { CloseNewTriangleLeftMiddle(i, AFWalker, NextLeftVertex);
      TInd = NextLeftVertex->TriangleIndex;
      AFLow->DeleteElement(AFWalker, NextTriangPosition);
      WorkFlag = 1;
      CleanRP = 1;
    }

    if (WorkFlag)
    { CheckDT(NextTriangPosition, TInd);
      CheckDT(NextTriangPosition, TriangArray[NextTriangPosition].ki);

      NextTriangPosition++;
      AFWalker = NextLeftVertex;
      NextLeftVertex = AFLow->ReturnNextLeft();
    }
  }
  return AFWalker;
}




long SetTIndex(long T,  long v1, long v2)
// returns triangle wich have border edge v1, v2;
// at first it starts with triangle T. It returns the index of 
// that triangle
{ if (TriangArray[T].BorderTriangleFlag)
  { if ( ((TriangArray[T].i == v1) && (TriangArray[T].j == v2)) ||
         ((TriangArray[T].i == v2) && (TriangArray[T].j == v1)))
      return T;
    
    if ( ((TriangArray[T].j == v1) && (TriangArray[T].k == v2)) ||
         ((TriangArray[T].j == v2) && (TriangArray[T].k == v1)))
      return T;

    if ( ((TriangArray[T].k == v1) && (TriangArray[T].i == v2)) ||
         ((TriangArray[T].k == v2) && (TriangArray[T].i == v1)))
      return T;
  }

  T--;
  while (T >= 0)
  { if (TriangArray[T].BorderTriangleFlag)
    { if ( ((TriangArray[T].i == v1) && (TriangArray[T].j == v2)) ||
           ((TriangArray[T].i == v2) && (TriangArray[T].j == v1)))
        return T;
    
      if ( ((TriangArray[T].j == v1) && (TriangArray[T].k == v2)) ||
           ((TriangArray[T].j == v2) && (TriangArray[T].k == v1)))
        return T;

      if ( ((TriangArray[T].k == v1) && (TriangArray[T].i == v2)) ||
           ((TriangArray[T].k == v2) && (TriangArray[T].i == v1)))
        return T;
    }
    T = T-1;
  }

  return 0;
}

int SolveLeftCase(long i)
// in this case, vertical projection of the inserted point i
// misses the advancing front from the left side.
{ double x = Vertices[i].x;
  double y = Vertices[i].y;
  
  DoubleLinkedList* L1 = AFLow->ReturnLeft();
  DoubleLinkedList* L2 = AFLow->ReturnNextRight();

// added on May 30
  int DeleteLeftFlag = 1;

  while (Equal(L1->key, L2->key) && Equal(L1->key,x))
  { L1 = L2;
    L2 = AFLow->ReturnNextRight();
    DeleteLeftFlag = 0;
    if (L2 == NULL) 
      return 0;
  }

// end of addtions

  long StoreVertex = L1->VertexNo;
  long StoreTIndex;

  if (SignedArea(x, y, L1->key, L1->y, L2->key, L2->y) > 0)
  { 
    InsertNewTriangle(i, L1, L2);
    long TInd = L1->TriangleIndex;
    if (DeleteLeftFlag)
      AFLow->DeleteLeft();
    if (AFLow->AppendAtTheLeft(x,y,i,NextTriangPosition) == 0)
      return 0;
    StoreTIndex = SetTIndex(NextTriangPosition, i, StoreVertex);

    NextTriangPosition++;

    GoRight(x, y, i, L2);
  }
  else
  { StoreTIndex = NOINDEX; 
    if (AFLow->AppendAtTheLeft(x,y,i,NextTriangPosition) == 0)
      return 0;
  }

  DoubleLinkedList* AFWalker = L2 = ReturnDCLLeft();
  DoubleLinkedList* NextRightVertex;
  if ((NextRightVertex = L2->RPointer) != NULL)
  { 

    while ((NextRightVertex !=  NULL) && (SignedArea(x, y, 
                            NextRightVertex->key,
                            NextRightVertex->y,
                            AFWalker->key, // x
                            AFWalker->y)) > 0  )
    { // new triangles have to be added
      CloseNewTriangleRightDown(i, StoreTIndex, AFWalker, NextRightVertex);
      StoreTIndex = NextTriangPosition;
      NextTriangPosition++;

      DeleteFromLeft(AFWalker);

      AFWalker = NextRightVertex;
      NextRightVertex = NextRightVertex->RPointer;

    }   
  }
  AFHighAppendLeft(x, y, i, StoreTIndex);
  return 1;
}



int SolveRightCase(long i)
// vertical projection of the inserted point i
// misses the advancing front from the right side.
{ double x = Vertices[i].x;
  double y = Vertices[i].y;
  
  DoubleLinkedList* L1 = AFLow->ReturnRight();
  DoubleLinkedList* L2 = AFLow->ReturnNextLeft();
  long ti = L2->TriangleIndex;
  long StoreVertex = L1->VertexNo;

  long StoreTIndex;

  if (SignedArea(x, y, L1->key, L1->y, L2->key, L2->y) < 0)
  { InsertNewTriangle(i, L2, L1);
    long TInd = L2->TriangleIndex;
    AFLow->DeleteRight();
    if (AFLow->AppendAtTheRight(x,y,i,NextTriangPosition) == 0)
      return 0;

    StoreTIndex = SetTIndex(NextTriangPosition, i, StoreVertex);

    NextTriangPosition++;

    GoLeft(x, y, i, L2);
  }
  else
  { StoreTIndex = NOINDEX;
    if (AFLow->AppendAtTheRight(x,y,i,NextTriangPosition) == 0)
      return 0;
  }

  DoubleLinkedList* AFWalker = L2 = ReturnDCLRight();
  DoubleLinkedList* NextLeftVertex;

  if ((NextLeftVertex = L2->LPointer) != NULL)
  { 

    while ((NextLeftVertex !=  NULL) && (SignedArea(x, y, 
                            NextLeftVertex->key,
                            NextLeftVertex->y,
                            AFWalker->key, // x
                            AFWalker->y)) < 0  )
    { // new triangles have to be added
      CloseNewTriangleLeftDown(i, StoreTIndex, NextLeftVertex, AFWalker);
      StoreTIndex = NextTriangPosition;
      NextTriangPosition++;

      DeleteFromRight(AFWalker);

      AFWalker = NextLeftVertex;
      NextLeftVertex = NextLeftVertex->LPointer;
    }

  }
  AFHighAppendRight(x, y, i, StoreTIndex);
  return 1;
}



void  PlotTrianglesSweep(HDC hdc, double BoxXMin, double BoxYMin, 
                                  double BoxXMax, double BoxYMax, 
                                  int right, int bottom)
{ long ii;
  double x1, x2, x3, y1, y2, y3;
  int xi1, yi1, xi2, yi2, xi3, yi3;



  POINT plxy[5];

  HPEN hpen, hpenOld;
  HBRUSH HBrush = CreateSolidBrush(RGB(0,255,0));
  (HBRUSH)SelectObject(hdc, HBrush);

  hpen = CreatePen(PS_SOLID, 1, RGB(255,0,0));
  hpenOld = (HPEN)SelectObject(hdc, hpen);

  long i, j, k;
//  NextTriangPosition--;
  for (ii = 0; ii < NextTriangPosition; ii++)
  { i = TriangArray[ii].i;
    j = TriangArray[ii].j;
    k = TriangArray[ii].k;

    x1 = Vertices[i].x;
    y1 = Vertices[i].y;

    x2 = Vertices[j].x;
    y2 = Vertices[j].y;

    x3 = Vertices[k].x;
    y3 = Vertices[k].y;

    xi1 = FromRealToPixelX(x1, BoxXMin, BoxXMax - BoxXMin, right);
    yi1 = bottom - FromRealToPixelY(y1, BoxYMin, BoxYMax - BoxYMin, bottom);
    xi2 = FromRealToPixelX(x2, BoxXMin, BoxXMax - BoxXMin, right);
    yi2 = bottom - FromRealToPixelY(y2, BoxYMin, BoxYMax - BoxYMin, bottom);
    xi3 = FromRealToPixelX(x3, BoxXMin, BoxXMax - BoxXMin, right);
    yi3 = bottom - FromRealToPixelY(y3, BoxYMin, BoxYMax - BoxYMin, bottom);

    plxy[0].x = xi1;   plxy[0].y = yi1;
    plxy[1].x = xi2;   plxy[1].y = yi2;
    plxy[2].x = xi3;   plxy[2].y = yi3;

//    Polygon(hdc, plxy, 3);
 
    MoveToEx(hdc, xi1, yi1, NULL);
    LineTo(hdc, xi2, yi2);
    LineTo(hdc, xi3, yi3);
    LineTo(hdc, xi1, yi1);
  }

  SelectObject(hdc, hpenOld);
  DeleteObject(hpen);
}


void SaveTriangle(char* FName)
{
  FILE* f = fopen(FName, "wt");
  if (f != NULL)
  { int i;
    fprintf(f, "%d \n", NoOfVertices);
    for (i = 0; i < NoOfVertices; i++)
    { fprintf(f, "%d, %lf, %lf \n", Vertices[i].Index, Vertices[i].x, Vertices[i].y);
    }
    fprintf(f, "%d \n", NextTriangPosition);
    for (i = 0; i < NextTriangPosition; i++)
    { fprintf(f, "%d, %d, %d \n", TriangArray[i].i, TriangArray[i].j, TriangArray[i].k);
    }


  }
}


int PrepareTriangles(long t1, long t2)
// this function rotates the triangles t1 and t2 in that way 
// that Delaunay criterion can be applied.
{
  while (( TriangArray[t1].k == TriangArray[t2].i) || 
         ( TriangArray[t1].k == TriangArray[t2].j) || 
         ( TriangArray[t1].k == TriangArray[t2].k))
    ShiftVerticesRight(&TriangArray[t1]);  


  if ((TriangArray[t1].i == TriangArray[t2].j) && 
      (TriangArray[t1].j == TriangArray[t2].i))
    return 1;

  ShiftVerticesRight(&TriangArray[t2]);

  if ((TriangArray[t1].i == TriangArray[t2].j) && 
      (TriangArray[t1].j == TriangArray[t2].i))
    return 1;

  ShiftVerticesRight(&TriangArray[t2]);
  if ((TriangArray[t1].i == TriangArray[t2].j) && 
      (TriangArray[t1].j == TriangArray[t2].i))
    return 1;
  return 0;
}


void SHRVertices(BaseTriangle* t)
// shift vertices right
{ long a;

  a = t->k;
  t->k = t->j;
  t->j = t->i;
  t->i = a;

  a = t->ki;
  t->ki = t->jk;
  t->jk = t->ij;
  t->ij = a;
}


void ReplaceNeighbour1(long TIndex, long OldT, long NewNeighbour)
{ if (TriangArray[TIndex].ij == OldT) 
    TriangArray[TIndex].ij = NewNeighbour;
  else
  { if (TriangArray[TIndex].jk == OldT)
      TriangArray[TIndex].jk = NewNeighbour;
    else
    { if (TriangArray[TIndex].ki == OldT)
       TriangArray[TIndex].ki = NewNeighbour;
    }
  }
}


void CheckDT(long t1, long t2)
// Triangles t1 and t2 are checked for Delaunay criterium and 
// legalised in neccessary
{ 

  if ((t1 != NOINDEX) && (t2 != NOINDEX))
   if (t1 != t2)
    { long i1, j1, k1, k2;

      if (PrepareTriangles(t1, t2))
      {
        if (PrepareTriangles(t2, t1))
        {


          i1 = TriangArray[t1].i;
          j1 = TriangArray[t1].j;
          k1 = TriangArray[t1].k; 
          k2 = TriangArray[t2].k;     
          C[0][0] = Vertices[i1].x - Vertices[k2].x;
          C[0][1] = Vertices[i1].y - Vertices[k2].y;
          C[0][2] = C[0][0] * C[0][0] + C[0][1] * C[0][1];

          C[1][0] = Vertices[j1].x - Vertices[k2].x;
          C[1][1] = Vertices[j1].y - Vertices[k2].y;
          C[1][2] = C[1][0] * C[1][0] + C[1][1] * C[1][1];

          C[2][0] = Vertices[k1].x - Vertices[k2].x;
          C[2][1] = Vertices[k1].y - Vertices[k2].y;
          C[2][2] = C[2][0] * C[2][0] + C[2][1] * C[2][1];

          double D =   C[0][0] * (C[1][1] * C[2][2] - C[1][2] * C[2][1])
                     + C[0][1] * (C[1][2] * C[2][0] - C[1][0] * C[2][2])
                     + C[0][2] * (C[1][0] * C[2][1] - C[1][1] * C[2][0]);

 
          if (Equal(D, 0)) goto w;

          if(D < 0) 
          { 
            SwapCounter++;

            long n_ij1 = TriangArray[t1].ij; 
            long n_jk1 = TriangArray[t1].jk;
            long n_ki1 = TriangArray[t1].ki;

            long n_ij2 = TriangArray[t2].ij; 
            long n_jk2 = TriangArray[t2].jk;
            long n_ki2 = TriangArray[t2].ki;

            // new triangles are at the location of the old triangles
            TriangArray[t1].i = k1;     
            TriangArray[t1].j = i1;   
            TriangArray[t1].k = k2;
      
            TriangArray[t2].i = k1;     
            TriangArray[t2].j = k2;   
            TriangArray[t2].k = j1;

            // update the neighbours
            TriangArray[t1].ij = n_ki1;
            TriangArray[t1].jk = n_jk2;
            TriangArray[t1].ki = t2;

            TriangArray[t2].ij = t1;
            TriangArray[t2].jk = n_ki2;
            TriangArray[t2].ki = n_jk1;

            if (n_jk2 != NOINDEX)
              ReplaceNeighbour1(n_jk2, t2, t1);      

            if (n_jk1 != NOINDEX)
              ReplaceNeighbour1(n_jk1, t1, t2);
 
            CheckBorderFlag(t1);
            CheckBorderFlag(t2);
            if (TriangArray[t1].BorderTriangleFlag)
              UpdateTriangleIndex(t1);

              if (TriangArray[t2].BorderTriangleFlag)
                UpdateTriangleIndex(t2);

              CheckBorderFlag(t1);
              CheckBorderFlag(t2);
  
            CheckDT(t1, n_jk2);
            CheckDT(t2, n_ki2);
            CheckDT(t1, n_ki1);
            CheckDT(t2, n_jk1);
          }
        }
      }
    }
w:;
}

void UpdateTriangleIndex(int ti)

//  Azuriranje AFLow je dovoljeno le glede na rob;
// prav tako je smiselno primerjati le indekse oglisc.
{ int DoneFlag = 0;

  if (TriangArray[ti].ij == NOINDEX)
  { DoneFlag = AFLow->UpdateTIndex(Vertices[TriangArray[ti].i].x,
                                   TriangArray[ti].i, 
                                   TriangArray[ti].j, ti);
    if (DoneFlag == 0)
      UpdateTIndexLow(TriangArray[ti].i, TriangArray[ti].j, ti);
  }

  if (TriangArray[ti].jk == NOINDEX)
  { DoneFlag = AFLow->UpdateTIndex(Vertices[TriangArray[ti].j].x,
                                   TriangArray[ti].j, 
                                   TriangArray[ti].k, ti);
    if (DoneFlag == 0)
      UpdateTIndexLow(TriangArray[ti].j, TriangArray[ti].k, ti);
  }

  if (TriangArray[ti].ki == NOINDEX)
  { DoneFlag = AFLow->UpdateTIndex(Vertices[TriangArray[ti].k].x,
                                   TriangArray[ti].k, 
                                   TriangArray[ti].i, ti);
    if (DoneFlag == 0)
      UpdateTIndexLow(TriangArray[ti].k, TriangArray[ti].i, ti);
  }
}

long ReturnSwapCounter() {return SwapCounter;}
long ReturnNoOfTriangle() {return NextTriangPosition;}

void PrintTriangulation()
{ int i, bf;

  FILE* f = fopen("TRIANGLES.TXT", "wt");
  for (i = 0; i < NextTriangPosition; i++)
  { if (TriangArray[i].BorderTriangleFlag) bf = 1;
    else bf = 0;

    fprintf(f, "%d. i:%ld j:%ld k:%ld ij:%ld jk:%ld ki:%ld BTF:%d\n",
                i, TriangArray[i].i, TriangArray[i].j, TriangArray[i].k, 
               TriangArray[i].ij, TriangArray[i].jk, TriangArray[i].ki,bf);
  }

  fclose(f);
}



void RemoveLongTriangles(HDC hdc, double dT,
                                  double BoxXMin, double BoxYMin, 
                                  double BoxXMax, double BoxYMax, 
                                  int right, int bottom)
{
 long ii;
  double x1, x2, x3, y1, y2, y3;
  int xi1, yi1, xi2, yi2, xi3, yi3;

  double dd = dT * dT;

  POINT plxy[5];

  HPEN hpen, hpenOld;
  HBRUSH HBrush = CreateSolidBrush(RGB(0,0,0));
  (HBRUSH)SelectObject(hdc, HBrush);

  hpen = CreatePen(PS_SOLID, 1, RGB(0,0,0));
  hpenOld = (HPEN)SelectObject(hdc, hpen);

  long i, j, k;
  double a,b,c;
//  NextTriangPosition--;
  for (ii = 0; ii < NextTriangPosition; ii++)
  { i = TriangArray[ii].i;
    j = TriangArray[ii].j;
    k = TriangArray[ii].k;

    x1 = Vertices[i].x;
    y1 = Vertices[i].y;

    x2 = Vertices[j].x;
    y2 = Vertices[j].y;

    x3 = Vertices[k].x;
    y3 = Vertices[k].y;

    a = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
    b = (x3 - x1) * (x3 - x1) + (y3 - y1) * (y3 - y1);
    c = (x2 - x3) * (x2 - x3) + (y2 - y3) * (y2 - y3);

    if ((a < dd) && (b < dd) && (c < dd))
    {
    xi1 = FromRealToPixelX(x1, BoxXMin, BoxXMax - BoxXMin, right);
    yi1 = bottom - FromRealToPixelY(y1, BoxYMin, BoxYMax - BoxYMin, bottom);
    xi2 = FromRealToPixelX(x2, BoxXMin, BoxXMax - BoxXMin, right);
    yi2 = bottom - FromRealToPixelY(y2, BoxYMin, BoxYMax - BoxYMin, bottom);
    xi3 = FromRealToPixelX(x3, BoxXMin, BoxXMax - BoxXMin, right);
    yi3 = bottom - FromRealToPixelY(y3, BoxYMin, BoxYMax - BoxYMin, bottom);

    plxy[0].x = xi1;   plxy[0].y = yi1;
    plxy[1].x = xi2;   plxy[1].y = yi2;
    plxy[2].x = xi3;   plxy[2].y = yi3;

//    Polygon(hdc, plxy, 3);
 
    MoveToEx(hdc, xi1, yi1, NULL);
    LineTo(hdc, xi2, yi2);
    LineTo(hdc, xi3, yi3);
    LineTo(hdc, xi1, yi1);
    }
  }

  SelectObject(hdc, hpenOld);
  DeleteObject(hpen);
}
//  PlotTriangulatedPoSweep, ints(hdc, BoxXMin, BoxYMin, BoxXMax, BoxYMax, r.right, r.bottom, NoOfPoints);
